#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import platform
import functools
import typing as t
from lxml import etree
from pathlib import (
	PureWindowsPath,
	PurePosixPath,
	Path
)


__all__ = [
	"search_upwards",
	"resolve_relpath",
	"read_cached_file",
	"read_cached_xml_file"
]


@functools.lru_cache(maxsize=None)
def search_upwards(
		for_path: t.Union[Path, str],
		from_path: t.Union[Path, str] = __file__
) -> Path:
	current = Path(from_path).resolve()
	while current != current.parent:
		new_path = current / for_path
		if new_path.exists():
			return new_path
		current = current.parent
	raise FileNotFoundError(for_path)


@functools.lru_cache(maxsize=None)
def resolve_relpath(path: t.Union[str, Path] = '.') -> Path:
	path_cls = PurePosixPath
	if platform.system().lower() == "windows":  # pragma: no cover
		path_cls = PureWindowsPath

	pure_path = path_cls(path)
	if pure_path.is_absolute():
		return Path(pure_path)
	return (Path.cwd() / path).resolve()


@functools.lru_cache(maxsize=None)
def read_cached_file(
		path: t.Union[str, Path],
		binary: bool = False,
		as_lines: bool = False,
		encoding: str = "UTF-8"
) -> t.Union[str, bytes, t.List[str], t.List[bytes]]:
	path = resolve_relpath(path)
	mode = 'rb' if binary else 'r'
	args = {} if binary else dict(encoding=encoding)
	with path.open(mode, **args) as file:
		func = file.readlines if as_lines else file.read
		return func()


@functools.lru_cache(maxsize=None)
def read_cached_xml_file(
		path: t.Union[str, Path],
		return_as_etree: bool = False
) -> t.Union[str, etree.Element]:
	"""
	Reads and parses the contents of an XML file.
	:param path: Either an absolute or a relative path to the XML file.
		If the path is relative, it is evaluated from the current working directory of the active shell.
	:param return_as_etree: If the file content should be returned as an instance of etree.Element.
		By default, the XML content is returned as a string.
	:raises lxml.etree.LxmlError: If lxml library fails to validate the content of the XML file.
		This error type is the superclass of lxml library errors.
	:return: The XML file content as a string or an etree.Element.
	"""
	path = resolve_relpath(path)
	with path.open('rb') as file:
		contents = file.read()
		obj = etree.fromstring(contents)  # Validates XML structure
		if return_as_etree:
			return obj
		return contents.decode('utf-8')



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from pydantic_xml import BaseXmlModel, element, attr
from fastapi_xroad_soap.internal.constants import XSD
from .restrictions import (
    NumericTypeRestriction,
    StringTypeRestriction
)


__all__ = [
    "AnyXML",
    "Element",
    "Sequence",
    "Attribute",
    "ComplexType",
    "SimpleType",
    "Import",
    "Include",
    "Schema"
]


class AnyXML(BaseXmlModel, tag="any"):
    process_contents: str = attr(name="processContents", default="lax")


class Element(BaseXmlModel, tag="element", skip_empty=True):
    name: str = attr()
    type: str = attr()
    min_occurs: t.Union[str, None] = attr(name="minOccurs", default=None)
    max_occurs: t.Union[str, None] = attr(name="maxOccurs", default=None)


class Sequence(BaseXmlModel, tag="sequence"):
    elements: t.List[t.Union[Element, AnyXML]]


class Attribute(BaseXmlModel, tag="attribute"):
    name: str = attr()
    type: str = attr()


class ComplexType(BaseXmlModel, tag="complexType", skip_empty=True):
    name: str = attr()
    sequence: Sequence
    attribute: t.List[Attribute] = element(tag="attribute", default=None)


class SimpleType(BaseXmlModel, tag="simpleType"):
    name: str = attr()
    restriction: t.Union[
        NumericTypeRestriction,
        StringTypeRestriction
    ]


class Import(BaseXmlModel, tag="import"):
    namespace: str = attr()
    schema_loc: str = attr(name="schemaLocation")


class Include(BaseXmlModel, tag="include"):
    schema_loc: str = attr(name="schemaLocation")


class Schema(BaseXmlModel, tag="schema"):
    xmlns: str = attr(default=XSD)
    target_ns: str = attr(name="targetNamespace")
    imports: t.List[Import]
    includes: t.List[Include]
    elements: t.List[Element]
    complex_types: t.List[ComplexType]
    simple_types: t.List[SimpleType]



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
from enum import Enum


__all__ = [
    "XSD",
    "ENV_NSMAP",
    "XRO_NSMAP",
    "IDEN_NSMAP",
    "HEADER_NSMAP",
    "WSDL_NSMAP",
    "A8nType"
]


XSD = "http://www.w3.org/2001/XMLSchema"
ENV_NSMAP = {"soapenv": "http://schemas.xmlsoap.org/soap/envelope/"}
XRO_NSMAP = {"xro": "http://x-road.eu/xsd/xroad.xsd"}
IDEN_NSMAP = {"iden": "http://x-road.eu/xsd/identifiers"}
HEADER_NSMAP = {**ENV_NSMAP, **XRO_NSMAP, **IDEN_NSMAP}
WSDL_NSMAP = {
    "wsdl": "http://schemas.xmlsoap.org/wsdl/",
    "soap": "http://schemas.xmlsoap.org/wsdl/soap/",
    "wsi": "http://ws-i.org/profiles/basic/1.1/xsd",
    **XRO_NSMAP
}


class A8nType(Enum):
    LIST = "list"
    OPT = "optional"
    MAND = "mandatory"
    ABSENT = "absent"



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
from .internal.soap import SoapFault, ClientFault


__all__ = ["SoapFault", "ClientFault"]



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
import inflection
from abc import ABC, abstractmethod
from pydantic_xml import model, element
from ..constants import A8nType
from . import validators as vld

try:
	from .message_body import MessageBody
except ImportError:  # pragma: no cover
	MessageBody: t.TypeAlias = t.Any


__all__ = ["BaseElementSpec"]


class BaseElementSpec(ABC):
	tag: t.Optional[str]
	ns: t.Optional[str]
	nsmap: t.Optional[t.Dict[str, str]]
	min_occurs: int
	max_occurs: t.Union[int, t.Literal["unbounded"]]
	element_type: t.Type[MessageBody]
	internal_type: t.Optional[t.Type[MessageBody]]
	a8n_type: t.Union[A8nType, None]

	def __init__(
			self,
			element_type: t.Type[MessageBody],
			internal_type: t.Optional[t.Type[MessageBody]] = None,
			**kwargs
	) -> None:
		self.tag = kwargs.get("tag")
		self.ns = kwargs.get("ns")
		self.nsmap = kwargs.get("nsmap")
		self.min_occurs = kwargs.get("min_occurs") or 0
		self.max_occurs = kwargs.get("max_occurs") or "unbounded"
		self.element_type = element_type
		self.internal_type = internal_type
		self.a8n_type = None

	@abstractmethod
	def init_instantiated_data(self, data: t.List[t.Any]) -> t.List[t.Any]: ...

	@abstractmethod
	def init_deserialized_data(self, data: t.List[t.Any]) -> t.List[t.Any]: ...

	@property
	@abstractmethod
	def has_constraints(self) -> bool: ...

	@abstractmethod
	def wsdl_type_name(self, *, with_tns: bool = False) -> str: ...

	@property
	def default_wsdl_type_name(self) -> str:
		raise NotImplementedError

	def assemble_wsdl_type_name(self, signature: str, with_tns: bool) -> str:
		if not self.has_constraints:
			return self.default_wsdl_type_name
		ns = "tns:" if with_tns else ''
		cap_name = self.default_wsdl_type_name.capitalize()
		return f"{ns}Custom{cap_name}__{signature}"

	def get_element_a8n(self) -> t.Type[t.List[t.Any]]:
		return t.List[self.internal_type or self.element_type]

	def get_element(self, attr: str) -> model.XmlEntityInfo:
		return element(
			tag=self.tag or inflection.camelize(attr),
			ns=self.ns or '',
			nsmap=self.nsmap or dict(),
			default_factory=list
		)

	def set_a8n_type_from(self, a8n: t.Any, attr: str, cls_name: str) -> None:
		et = self.element_type
		vld.validate_a8n_args(a8n, attr, cls_name, et)
		if a8n in [A8nType.ABSENT, et, et.__name__]:
			self.a8n_type = A8nType.MAND
			return
		if origin := t.get_origin(a8n):
			if origin == list:
				self.a8n_type = A8nType.LIST
				return
			elif "Union" in origin.__name__:
				self.a8n_type = A8nType.OPT
				return
		raise TypeError(
			f"Unsupported annotation '{a8n}' for "
			f"class '{cls_name}' attribute '{attr}'."
		)



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from pydantic_xml import element
from ..constants import ENV_NSMAP
from ..base import MessageBody, MessageBodyType
from .header import XroadHeader


__all__ = [
	"GenericEnvelope",
	"GenericFault",
	"GenericBody",
	"AnyBody"
]


bases = [MessageBody, t.Generic[MessageBodyType]]


class GenericEnvelope(*bases, tag="Envelope", nsmap=ENV_NSMAP, search_mode='unordered'):
	header: t.Optional[XroadHeader] = element(tag="Header", default=None)
	body: t.Optional[MessageBodyType] = element(tag="Body", default=None)


class GenericFault(*bases, tag="Fault", ns="soapenv", nsmap=ENV_NSMAP):
	faultcode: str = element(tag="faultcode", ns='')
	faultstring: str = element(tag="faultstring", ns='')
	faultactor: t.Optional[str] = element(tag="faultactor", ns='', default=None)
	detail: MessageBodyType = element(tag="detail", ns='', default=None)


class GenericBody(*bases, tag="Body", nsmap=ENV_NSMAP):
	content: MessageBodyType


class AnyBody(MessageBody, tag="Body", nsmap=ENV_NSMAP):
	content: t.Optional[MessageBody] = element(default_factory=MessageBody)



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from fastapi import Response
from ..base import MessageBody
from ..uid_gen import UIDGenerator
from ..multipart import MultipartEncoder
from ..elements.models import SwaRefUtils
from ..envelope import (
	EnvelopeFactory,
	XroadHeader
)


__all__ = ["SoapResponse"]


class SoapResponse(Response):
	media_type = 'text/xml'

	def __init__(
			self,
			content: MessageBody,
			header: t.Optional[XroadHeader] = None,
			http_status_code: t.Optional[int] = 200
	) -> None:
		if files := SwaRefUtils.gather_specs_and_files(content)[1]:
			cid = UIDGenerator(mode="cid")
			for file in files:
				file.content_id = cid.generate()
			message = self.serialize(content, header)
			encoder = MultipartEncoder(message, files)
			xml_str = encoder.message
			http_headers = encoder.headers
		else:
			xml_str = self.serialize(content, header)
			http_headers = {"Content-Type": "text/xml; charset=UTF-8"}
		super().__init__(
			content=xml_str,
			headers=http_headers,
			status_code=http_status_code
		)

	@staticmethod
	def serialize(content: MessageBody, header: t.Optional[XroadHeader]) -> bytes:
		envelope = EnvelopeFactory[content.__class__](
			exclude_xroad_nsmap=header is None
		)
		return envelope.serialize(
			content=content,
			header=header,
			pretty_print=False,
			skip_empty=True
		)



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from .base_element_spec import BaseElementSpec


__all__ = ["dynamic_spec"]


def dynamic_spec(_type: t.Type, **_kwargs) -> t.Any:
	class DynamicSpec(BaseElementSpec):
		def __init__(self) -> None:
			super().__init__(element_type=_type, **_kwargs)

		def init_instantiated_data(self, data: t.List) -> t.List:
			return data

		def init_deserialized_data(self, data: t.List) -> t.List:
			return data

		@property
		def has_constraints(self) -> bool:
			return False

		def wsdl_type_name(self, *, with_tns: bool = False) -> str:
			if with_tns:
				return f"tns:{_type.__name__}"
			return _type.__name__

	return DynamicSpec()



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
import inflection
from ..base import MessageBody
from ..constants import WSDL_NSMAP
from ..soap.action import SoapAction
from ..elements.models import SwaRefSpec
from . import helpers, models as mod


__all__ = ["generate"]


def generate(actions: t.Dict[str, SoapAction], name: str, tns: str) -> bytes:
	wsdl_def = type(
		"WSDLDefinitions", (mod.WSDLDefinitions,),
		{}, nsmap={**WSDL_NSMAP, "tns": tns}
	)
	wsdl_data: str = wsdl_def(
		target_ns=tns,
		name=name,
		types=mod.WSDLTypes(
			service_schema=mod.Schema(
				target_ns=tns,
				includes=[],
				imports=_generate_imports(actions),
				elements=_generate_elements(actions),
				**_generate_types(actions)
			)
		),
		messages=_generate_messages(actions),
		port_type=_generate_port_type(actions),
		binding=_generate_binding(actions),
		service=mod.WSDLService(
			port=mod.WSDLPortBinding(
				address=mod.SOAPAddress()
			)
		)
	).to_xml(pretty_print=True).decode()
	return helpers.format_wsdl_definitions(wsdl_data)


def _generate_imports(actions: t.Dict[str, SoapAction]) -> t.List[mod.Import]:
	xro = "http://x-road.eu/xsd/xroad.xsd"
	imports = [mod.Import(
		namespace=xro,
		schema_loc=xro
	)]
	for action in actions.values():
		for model in [action.body_type, action.return_type]:
			if model is None:
				continue
			_add_swaref_import(model, imports)
	return imports


def _add_swaref_import(model: t.Type[MessageBody], imports: t.List[mod.Import]) -> None:
	wsi = "http://ws-i.org/profiles/basic/1.1/"
	for nested_model in model.nested_models():
		specs = nested_model.model_specs()
		for spec in specs.values():
			if not isinstance(spec, SwaRefSpec):
				continue
			imports.append(mod.Import(
				namespace=f"{wsi}xsd",
				schema_loc=f"{wsi}swaref.xsd"
			))
			return


def _generate_elements(actions: t.Dict[str, SoapAction]) -> t.List[mod.Element]:
	elements = {}
	for name, action in actions.items():
		for model in [action.body_type, action.return_type]:
			if model is None:
				continue
			wsdl_name = model.wsdl_name()
			if wsdl_name in elements:
				continue
			elements[wsdl_name] = mod.Element(
				name=wsdl_name,
				type=f"tns:{wsdl_name}"
			)
	return [
		*list(elements.values()),
		mod.Element(
			name="FaultResponse",
			type="tns:FaultResponse"
		)
	]


def _generate_types(actions: t.Dict[str, SoapAction]) -> t.Dict:
	complex_types, simple_types = helpers.gather_all_types(actions)
	return dict(
		complex_types=[
			*complex_types,
			mod.ComplexType(
				name="FaultResponse",
				sequence=mod.Sequence(
					elements=[
						mod.Element(name="faultcode", type="string"),
						mod.Element(name="faultstring", type="string"),
						mod.Element(name="faultactor", type="string"),
						mod.Element(name="detail", type="tns:FaultResponseDetail")
					]
				)
			),
			mod.ComplexType(
				name="FaultResponseDetail",
				sequence=mod.Sequence(
					elements=[mod.AnyXML()]
				)
			)
		],
		simple_types=simple_types
	)


def _generate_messages(actions: t.Dict[str, SoapAction]) -> t.List[mod.WSDLMessage]:
	messages = {}
	for action in actions.values():
		for model in [action.body_type, action.return_type]:
			if model is None:
				continue
			wsdl_name = model.wsdl_name()
			if wsdl_name in messages:
				continue
			messages[wsdl_name] = mod.WSDLMessage(
				name=wsdl_name,
				parts=[mod.WSDLPart(
					element=f"tns:{wsdl_name}"
				)]
			)
	return [
		*list(messages.values()),
		mod.WSDLMessage(
			name="FaultResponse",
			parts=[mod.WSDLPart(element="tns:FaultResponse", name="fault")]
		),
		mod.WSDLMessage(
			name="xroadHeader",
			parts=[
				mod.WSDLPart(element="xro:client", name="client"),
				mod.WSDLPart(element="xro:service", name="service"),
				mod.WSDLPart(element="xro:id", name="id"),
				mod.WSDLPart(element="xro:protocolVersion", name="protocolVersion"),
				mod.WSDLPart(element="xro:userId", name="userId")
			]
		)
	]


def _generate_port_type(actions: t.Dict[str, SoapAction]) -> mod.WSDLPortType:
	ops = []
	for name, action in actions.items():
		port_docs = port_input = port_output = None
		if action.description is not None:
			port_docs = mod.WSDLDocumentation(
				title=inflection.titleize(name),
				notes=action.description
			)
		if action.body_type is not None:
			wsdl_name = action.body_type.wsdl_name()
			port_input = mod.WSDLInputPort(
				message=f"tns:{wsdl_name}",
				name=wsdl_name
			)
		if action.return_type is not None:
			wsdl_name = action.return_type.wsdl_name()
			port_output = mod.WSDLOutputPort(
				message=f"tns:{wsdl_name}",
				name=wsdl_name
			)
		ops.append(mod.WSDLOperationPort(
			documentation=port_docs,
			input=port_input,
			output=port_output,
			name=name
		))
	return mod.WSDLPortType(operations=ops)


def _generate_binding(actions: t.Dict[str, SoapAction]) -> mod.WSDLBinding:
	return mod.WSDLBinding(
		binding=mod.SOAPBinding(),
		operations=[
			mod.WSDLOperationBinding(
				name=key,
				operation=mod.SOAPOperationBinding(
					soap_action=key
				)
			)
			for key in actions.keys()
		]
	)



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from fastapi.types import DecoratedCallable
from ..envelope import XroadHeader
from ..base import MessageBody
from .. import utils


__all__ = ["validate_annotations"]


def validate_annotations(name: str, func: DecoratedCallable) -> dict:
	anno = utils.get_annotations(func)
	for key, value in anno.items():
		if key == "body":
			_validate_a8n_body(key, value, name)
		elif key == "header":
			_validate_a8n_header(value, name)
		elif key == "return":
			_validate_a8n_return(value, name)
		else:
			raise ValueError(
				f"Parameter name '{key}' not allowed for SOAP action {name}."
				"\nOnly names 'body' and 'header' can be used for parameters."
			)
	return anno


def _validate_a8n_body(key: str, value: t.Any, name: str) -> None:
	if value == XroadHeader:
		raise TypeError(
			f"Cannot set the 'XroadHeader' class as a type annotation "
			f"to the 'body' parameter of the {name} SOAP action."
		)
	elif not isinstance(value, type) or not issubclass(value, MessageBody):
		raise TypeError(
			f"The annotation of the '{key}' parameter of the {name} "
			f"SOAP action must be a subclass of 'MessageBody'."
		)


def _validate_a8n_header(value: t.Any, name: str) -> None:
	if value != XroadHeader:
		raise ValueError(
			f"The annotation of the 'headers' parameter of the {name} "
			f"SOAP action must be the 'XroadHeaders' class."
		)


def _validate_a8n_return(value: t.Any, name: str) -> None:
	if value is None:
		return
	if not isinstance(value, type) or not issubclass(value, MessageBody):
		raise TypeError(
			f"Return type annotation of the {name} SOAP action "
			"must be either 'None' or a subclass of 'MessageBody'."
		)



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from pydantic_xml import BaseXmlModel, element, attr
from ..constants import HEADER_NSMAP


__all__ = ["XroadService", "XroadClient", "XroadHeader"]


class XroadService(BaseXmlModel, tag="service", ns="xro", nsmap=HEADER_NSMAP, search_mode='unordered'):
	object_type: str = attr(name="objectType", ns="iden", default="SERVICE")
	xroad_instance: str = element(tag="xRoadInstance", ns="iden")
	member_class: str = element(tag="memberClass", ns="iden")
	member_code: str = element(tag="memberCode", ns="iden")
	subsystem_code: t.Union[str, None] = element(tag="subsystemCode", ns="iden", default=None)
	service_code: str = element(tag="serviceCode", ns="iden")
	service_version: t.Union[str, None] = element(tag="serviceVersion", ns="iden", default=None)


class XroadClient(BaseXmlModel, tag="client", ns="xro", nsmap=HEADER_NSMAP, search_mode='unordered'):
	object_type: str = attr(name="objectType", ns="iden", default="SUBSYSTEM")
	xroad_instance: str = element(tag="xRoadInstance", ns="iden")
	member_class: str = element(tag="memberClass", ns="iden")
	member_code: str = element(tag="memberCode", ns="iden")
	subsystem_code: t.Union[str, None] = element(tag="subsystemCode", ns="iden", default=None)


class XroadHeader(BaseXmlModel, tag="Header", ns="soapenv", nsmap=HEADER_NSMAP, search_mode='unordered'):
	user_id: t.Union[str, None] = element(tag="userId", ns="xro", default=None)
	proto_ver: str = element(tag="protocolVersion", ns="xro")
	id: str = element(tag="id", ns="xro")
	service: t.Union[XroadService, None] = None
	client: XroadClient



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import base64
import secrets
import typing as t


__all__ = ["UIDGenerator"]


class UIDGenerator:
	def __init__(self, mode: t.Literal["cid", "key"]):
		"""
		Initialize a UIDGenerator instance with a specific operational mode.
		This class generates unique identifiers (UIDs) based on the selected mode,
		either 'cid' or 'key'. The 'cid' mode generates content identifiers with
		a hexadecimal format, while the 'key' mode produces more complex keys
		encoded in a modified base64 format.

		:param mode: Determines the type of UID to generate. "cid" mode generates
			UIDs prefixed with 'cid:' and uses a shorter token length. "key" mode
			generates UIDs modified from base64 encoding, with specific character
			replacements and a longer token length.
		"""
		self.tokens = list()
		self._mode = mode
		self._max_len = 14 if mode == "cid" else 18
		self._token_len = 10 if mode == "cid" else 12
		self._increment = 1 if mode == "cid" else 3
		self._compute_token = (
			self._compute_cid_token
			if mode == "cid" else
			self._compute_key_token
		)

	def generate(self, from_bytes: t.Optional[bytes] = None) -> str:
		"""
		Generate a unique identifier (UID) based on the previously set mode.
		This method can either generate a new UID using random bytes or transform
		provided bytes into a UID format depending on the mode.

		If 'from_bytes' is provided, the UID is directly computed from these
		bytes without generating new random bytes. If not provided, random bytes
		of a specific length are used.

		The method ensures that all generated UIDs are unique by maintaining
		a history of previously generated UIDs. If a UID collision occurs, it
		attempts to resolve this by increasing the token length and retrying
		a limited number of times.

		:param from_bytes: Optional bytes to use for generating the UID.
			If not provided, random bytes are generated.
		:return: The generated unique identifier as a string.
		:raises RuntimeError: Raised if the UID generation exceeds
			the internally hardcoded maximum token length limit.
		"""
		exists_counter = 0
		while True:
			if self._token_len > self._max_len:
				raise RuntimeError("UIDGenerator out of bounds")

			token_bytes = from_bytes
			if not isinstance(from_bytes, bytes):
				token_bytes = secrets.token_bytes(self._token_len)
			token = self._compute_token(token_bytes)

			if isinstance(from_bytes, bytes):
				return token
			elif token not in self.tokens:
				self.tokens.append(token)
				return token
			elif exists_counter < 10:
				exists_counter += 1
			else:
				self._token_len += self._increment
				exists_counter = 0

	@staticmethod
	def _compute_cid_token(from_bytes: bytes = None) -> str:
		return f"cid:{from_bytes.hex()}"

	@staticmethod
	def _compute_key_token(from_bytes: bytes = None) -> str:
		token = base64.b64encode(from_bytes).decode()
		table = str.maketrans("+/=", "ABC")
		return token.translate(table)



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
from __future__ import annotations
import typing as t
from pydantic_xml import model
from pydantic import fields
from pydantic import (
	PrivateAttr,
	ValidationInfo,
	model_validator
)
from .. import utils
from ..constants import A8nType
from .base_element_spec import BaseElementSpec
from .composite_meta import CompositeMeta
from .dynamic_spec import dynamic_spec
from . import validators as vld


__all__ = ["MessageBody", "MessageBodyType"]


class MessageBody(model.BaseXmlModel, metaclass=CompositeMeta, search_mode='unordered', skip_empty=True):
	_element_specs: t.Dict[str, BaseElementSpec] = PrivateAttr(default_factory=dict)
	_T = t.TypeVar('_T', bound="MessageBody")

	@classmethod
	def wsdl_name(cls) -> str:
		return cls.__xml_tag__ or cls.__name__

	@classmethod
	def Element(
			cls: t.Type[_T],
			*,
			tag: t.Optional[str] = None,
			ns: t.Optional[str] = None,
			nsmap: t.Optional[t.Dict[str, str]] = None,
			min_occurs: int = None,
			max_occurs: t.Union[int, t.Literal["unbounded"]] = None
	) -> _T:
		kwargs = {k: v for k, v in locals().items() if v != cls}
		return t.cast(MessageBody, dynamic_spec(cls, **kwargs))

	@classmethod
	def model_specs(cls) -> t.Dict[str, BaseElementSpec]:
		privates = getattr(cls, "__private_attributes__", {})
		attr: t.Union[fields.ModelPrivateAttr, None] = privates.get("_element_specs")
		return attr.get_default() if attr is not None else dict()

	@classmethod
	def nested_models(cls) -> t.List[t.Type["MessageBody"]]:
		excluded_models = ["SwaRefInternal"]
		models = []
		a8ns = getattr(cls, '__annotations__', {})
		for key, value in a8ns.items():
			origin = t.get_origin(value)
			if origin is None:
				continue
			args = t.get_args(value)
			args_len = len(args)
			opt_union = args_len == 2 and type(None) in args
			if not opt_union and args_len != 1:
				raise ValueError(
					f"Invalid type annotation arguments '{args}' "
					f"for class {cls.__name__} attribute '{key}'"
				)
			value = [a for a in args if a is not None][0]
			if (
				type(value) is not CompositeMeta
				or value.__name__ in excluded_models
			):
				continue
			models.extend(value.nested_models())
		models.append(cls)
		return models

	# noinspection PyNestedDecorators
	@model_validator(mode="before")
	@classmethod
	def _validate_before(cls, data: t.Dict[str, t.Any], info: ValidationInfo) -> t.Any:
		# Used for validating model instantiation in user code
		if utils.is_incoming_request(info) or not isinstance(data, dict):
			return data

		for attr, spec in cls.model_specs().items():
			expected = spec.internal_type or spec.element_type
			value = data.get(attr)

			if spec.a8n_type == A8nType.LIST:
				vld.validate_list_items(attr, value, spec)
				spec.init_instantiated_data(value)
				continue
			elif spec.a8n_type == A8nType.OPT and value is None:
				data[attr] = []
				continue
			elif spec.a8n_type == A8nType.MAND and attr not in data:
				raise ValueError(f"argument '{attr}' is missing")
			elif not isinstance(value, expected):
				arg_type = type(value).__name__
				raise ValueError(f"unexpected type '{arg_type}' for argument '{attr}'")
			data[attr] = [value]
			spec.init_instantiated_data(data[attr])
		return data

	@model_validator(mode="after")
	def _validate_after(self, info: ValidationInfo) -> MessageBody:
		# Used for validating deserialized incoming requests
		if not utils.is_incoming_request(info):
			return self

		specs = getattr(self, "_element_specs", None)
		if not specs:
			return self

		for attr, value in vars(self).items():
			spec = specs.get(attr)  # type: BaseElementSpec
			if spec is None:
				continue
			elif not isinstance(value, list):
				value = [value]

			vld.validate_opt_mand_a8n(attr, value, spec)
			if spec.a8n_type == A8nType.LIST:
				vld.validate_list_items(attr, value, spec)
				spec.init_deserialized_data(value)
				continue

			spec.init_deserialized_data(value)
			value = value[0] if len(value) == 1 else None
			setattr(self, attr, value)
		return self


MessageBodyType = t.TypeVar("MessageBodyType", bound=MessageBody)



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from pathlib import Path
from fastapi.types import DecoratedCallable
from starlette.types import Scope, Receive, Send, Lifespan
from fastapi import Request, Response, FastAPI as ASGIApp
from ..storage import GlobalWeakStorage
from .action import SoapAction


FuncOrCoro = t.Union[t.Callable[..., t.Any], t.Awaitable[t.Any]]


class FastAPI:
	def __call__(self, scope: Scope, receive: Receive, send: Send) -> t.Awaitable[None]: ...


class SoapService(FastAPI):
	_name: str
	_tns: str
	_wsdl_response: t.Union[Response, None]
	_wsdl_override: t.Optional[t.Union[str, Path]]
	_fault_callback: t.Optional[t.Callable[[Request, Exception], None]] = None
	_hide_ise_cause: bool = False
	_storage: GlobalWeakStorage
	_actions: dict[str, SoapAction]

	def __init__(
			self,
			*,
			name: str = "SoapService",
			path: str = "/service",
			this_namespace: str = "https://example.org",
			wsdl_override: t.Optional[t.Union[str, Path]] = None,
			lifespan: t.Optional[Lifespan[FastAPI]] = None,
			fault_callback: t.Optional[t.Callable[[Request, Exception], None]] = None,
			hide_ise_cause: bool = False
	) -> None: ...

	async def _soap_middleware(self, http_request: Request, _: t.Callable) -> t.Optional[Response]: ...

	def _determine_action(self, http_request: Request) -> SoapAction: ...

	@staticmethod
	async def _await_or_call(func: FuncOrCoro, *args, **kwargs) -> t.Any: ...

	def add_action(
			self,
			name: str,
			handler: t.Callable[..., t.Any],
			description: t.Optional[str] = None
	) -> None: ...

	def action(
			self,
			name: str,
			description: t.Optional[str] = None
	) -> t.Callable[[DecoratedCallable], DecoratedCallable]: ...

	def regenerate_wsdl(self, *, force: bool = False) -> None: ...



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import re
import typing as t
import inflection
from ..soap.action import SoapAction
from ..constants import A8nType
from ..elements import NumericTypeSpec
from ..elements import StringTypeSpec
from ..base import (
	MessageBody,
	CompositeMeta,
	BaseElementSpec
)
from . import models as mod


__all__ = [
	"format_wsdl_definitions",
	"gather_all_types"
]


_AllTypes = t.Tuple[
	t.List[mod.ComplexType],
	t.List[mod.SimpleType]
]


def format_wsdl_definitions(wsdl_string: str) -> bytes:
	return re.sub(
		pattern=r'(<wsdl:definitions[^>]*>)',
		repl=lambda match: match.group(1).replace(' ', '\n  '),
		string=wsdl_string,
		count=1
	).encode()


def gather_all_types(actions: t.Dict[str, SoapAction]) -> _AllTypes:
	simple_types: t.Dict[str, mod.SimpleType] = {}
	complex_types: t.List[mod.ComplexType] = []

	for model in _gather_models(actions):
		elements = []
		for name, spec in model.model_specs().items():
			if type(spec.element_type) is not CompositeMeta:
				_add_simple_type(spec, simple_types)
			elements.append(mod.Element(
				name=spec.tag or inflection.camelize(name),
				type=spec.wsdl_type_name(with_tns=True),
				max_occurs=(
					None if
					spec.a8n_type != A8nType.LIST
					else str(spec.max_occurs)
				),
				min_occurs={
					A8nType.MAND: None,
					A8nType.LIST: str(spec.min_occurs),
					A8nType.OPT: "0"
				}[spec.a8n_type]
			))
		complex_types.append(mod.ComplexType(
			name=model.wsdl_name(),
			sequence=mod.Sequence(
				elements=elements
			)
		))
	return complex_types, list(simple_types.values())


def _gather_models(actions: t.Dict[str, SoapAction]) -> t.List[t.Type["MessageBody"]]:
	models = []
	for action in actions.values():
		for model in [action.body_type, action.return_type]:
			if model is None:
				continue
			nested = model.nested_models()
			models.extend(nested)
	return list(set(models))


def _add_simple_type(spec: BaseElementSpec, simple_types: t.Dict[str, mod.SimpleType]) -> None:
	if spec.has_constraints:
		func = (
			_create_string_simple_type if
			isinstance(spec, StringTypeSpec)
			else _create_numeric_simple_type
		)
		st: mod.SimpleType = func(spec)
		if st.name not in simple_types:
			simple_types[st.name] = st


def _create_string_simple_type(spec: StringTypeSpec) -> mod.SimpleType:
	return mod.SimpleType(
		name=spec.wsdl_type_name(),
		restriction=mod.StringTypeRestriction(
			base=spec.default_wsdl_type_name,
			length=(
				None if spec.length is None else
				mod.Length(value=str(spec.length))
			),
			min_length=(
				None if spec.min_length is None or spec.length is not None else
				mod.MinLength(value=str(spec.min_length))
			),
			max_length=(
				None if spec.max_length is None or spec.length is not None else
				mod.MaxLength(value=str(spec.max_length))
			),
			whitespace=(
				None if spec.whitespace == "preserve" else
				mod.WhiteSpace(value=spec.whitespace)
			),
			pattern=(
				None if spec.pattern is None else
				mod.RegexPattern(value=spec.pattern)
			),
			enumerations=(
				None if spec.enumerations is None else [
					mod.Enumeration(value=str(e.value))
					for e in spec.enumerations
				]
			)
		)
	)


def _create_numeric_simple_type(spec: NumericTypeSpec) -> mod.SimpleType:
	return mod.SimpleType(
		name=spec.wsdl_type_name(),
		restriction=mod.NumericTypeRestriction(
			base=spec.default_wsdl_type_name,
			min_inclusive=(
				None if spec.min_value is None else
				mod.MinInclusive(value=str(spec.min_value))
			),
			max_inclusive=(
				None if spec.max_value is None else
				mod.MaxInclusive(value=str(spec.max_value))
			),
			total_digits=(
				None if spec.total_digits is None else
				mod.TotalDigits(value=str(spec.total_digits))
			),
			pattern=(
				None if spec.pattern is None else
				mod.RegexPattern(value=spec.pattern)
			),
			enumerations=(
				None if spec.enumerations is None else [
					mod.Enumeration(value=str(e.value))
					for e in spec.enumerations
				]
			)
		)
	)



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from enum import Enum
from ..string_type_spec import StringTypeSpec


__all__ = ["StringSpec", "String"]


class StringSpec(StringTypeSpec):
	def __init__(self, **kwargs) -> None:
		super().__init__(element_type=str, **kwargs)

	@property
	def default_wsdl_type_name(self) -> str:
		return "string"


class String:
	def __new__(
			cls,
			*,
			tag: t.Optional[str] = None,
			ns: t.Optional[str] = None,
			nsmap: t.Optional[t.Dict[str, str]] = None,
			min_occurs: int = None,
			max_occurs: t.Union[int, t.Literal["unbounded"]] = None,
			length: t.Optional[int] = None,
			min_length: t.Optional[int] = None,
			max_length: t.Optional[int] = None,
			enumerations: t.Optional[t.Type[Enum]] = None,
			pattern: t.Optional[str] = None,
			whitespace: t.Literal["preserve", "replace", "collapse"] = "preserve"
	) -> str:
		kwargs = {k: v for k, v in locals().items() if v != cls}
		return t.cast(str, StringSpec(**kwargs))



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
from .internal.elements.models import (
	Boolean,
	Date,
	DateTime,
	Float,
	Integer,
	NetRes,
	String,
	SwaRef,
	Time
)


__all__ = [
	"Boolean",
	"Date",
	"DateTime",
	"Float",
	"Integer",
	"NetRes",
	"String",
	"SwaRef",
	"Time"
]



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from fastapi_xroad_soap.internal.base import BaseElementSpec
from ..validators import CommonValidators


__all__ = ["BooleanSpec", "Boolean"]


class BooleanSpec(BaseElementSpec, CommonValidators):
	def __init__(self, **kwargs) -> None:
		super().__init__(element_type=bool, **kwargs)

	def process(self, obj: t.Any) -> t.Any:
		self.validate_type(obj, self.element_type)
		return obj

	def init_instantiated_data(self, data: t.List[bool]) -> t.List[bool]:
		return [self.process(obj) for obj in data]

	def init_deserialized_data(self, data: t.List[bool]) -> t.List[bool]:
		return [self.process(obj) for obj in data]

	@property
	def has_constraints(self) -> bool:
		return False

	def wsdl_type_name(self, *, with_tns: bool = False) -> str:
		return "boolean"


class Boolean:
	def __new__(
			cls,
			*,
			tag: t.Optional[str] = None,
			ns: t.Optional[str] = None,
			nsmap: t.Optional[t.Dict[str, str]] = None,
			min_occurs: int = None,
			max_occurs: t.Union[int, t.Literal["unbounded"]] = None
	) -> bool:
		kwargs = {k: v for k, v in locals().items() if v != cls}
		return t.cast(bool, BooleanSpec(**kwargs))



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import functools
import typing as t
from abc import abstractmethod
from ..base import BaseElementSpec
from .. import utils
from .validators import (
	CommonValidators,
	StringValidators
)


__all__ = ["StringTypeSpec"]


class StringTypeSpec(BaseElementSpec, CommonValidators, StringValidators):
	def __init__(self, element_type: t.Any, **kwargs) -> None:
		self.length = kwargs.get("length", None)
		self.min_length = self.length or kwargs.get("min_length", None)
		self.max_length = self.length or kwargs.get("max_length", None)
		self.enumerations = kwargs.get("enumerations", None)
		self.pattern = kwargs.get("pattern", None)
		self.whitespace = (
			kwargs.get("whitespace", None)
			or t.cast(t.Literal, "preserve")
		)
		self.validate_enum_value_types(element_type)
		super().__init__(
			element_type=element_type,
			**kwargs
		)

	def process(self, obj: t.Any) -> t.Any:
		self.validate_type(obj, self.element_type)
		self.validate_string_length(obj)
		self.validate_pattern(obj)
		self.validate_enumerations(obj)
		return self.process_whitespace(obj)

	def init_instantiated_data(self, data: t.List[t.Any]) -> t.List[t.Any]:
		return [self.process(obj) for obj in data]

	def init_deserialized_data(self, data: t.List[t.Any]) -> t.List[t.Any]:
		return [self.process(obj) for obj in data]

	@property
	@functools.lru_cache(maxsize=1)
	def has_constraints(self) -> bool:
		return self.whitespace != 'preserve' or any(
			attr is not None for attr in [
				self.length,
				self.min_length,
				self.max_length,
				self.pattern,
				self.enumerations
			]
		)

	def wsdl_type_name(self, *, with_tns: bool = False) -> str:
		return self.assemble_wsdl_type_name(
			signature=utils.compute_signature(
				self.length,
				self.min_length,
				self.max_length,
				self.whitespace,
				self.enumerations,
				self.pattern
			),
			with_tns=with_tns
		)

	@property
	@abstractmethod
	def default_wsdl_type_name(self) -> str: ...



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from enum import Enum
from ..numeric_type_spec import NumericTypeSpec


__all__ = ["FloatSpec", "Float"]


class FloatSpec(NumericTypeSpec):
	def __init__(self, **kwargs) -> None:
		super().__init__(element_type=float, **kwargs)

	@property
	def default_wsdl_type_name(self) -> str:
		return "float"


class Float:
	def __new__(
			cls,
			*,
			tag: t.Optional[str] = None,
			ns: t.Optional[str] = None,
			nsmap: t.Optional[t.Dict[str, str]] = None,
			min_occurs: int = None,
			max_occurs: t.Union[int, t.Literal["unbounded"]] = None,
			min_value: t.Optional[float] = None,
			max_value: t.Optional[float] = None,
			enumerations: t.Optional[t.Type[Enum]] = None,
			pattern: t.Optional[str] = None
	) -> float:
		kwargs = {k: v for k, v in locals().items() if v != cls}
		return t.cast(float, FloatSpec(**kwargs))



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import re
import typing as t
from fastapi import Response
from pydantic import (
	BaseModel,
	ValidationError,
	field_validator
)
from ..base import MessageBody
from ..storage import GlobalWeakStorage
from ..constants import HEADER_NSMAP
from ..multipart import (
	MultipartDecoder,
	DecodedBodyPart
)
from ..envelope import (
	EnvelopeFactory,
	GenericEnvelope,
	XroadHeader
)
from .response import SoapResponse
from . import faults as f


__all__ = ["SoapAction"]


_ArgsFrom = t.List[t.Union[MessageBody, XroadHeader]]
_RespFrom = t.Union[Response, SoapResponse]


class SoapAction(BaseModel, arbitrary_types_allowed=True):
	name: str
	handler: t.Callable[..., t.Optional[MessageBody]]
	description: t.Optional[str]
	body_type: t.Optional[t.Type[MessageBody]]
	body_index: t.Optional[int]
	header_type: t.Optional[t.Type[XroadHeader]]
	header_index: t.Optional[int]
	return_type: t.Optional[t.Type[MessageBody]]
	storage: GlobalWeakStorage

	# noinspection PyNestedDecorators
	@field_validator("name")
	@classmethod
	def validate_name(cls, value: t.Any) -> str:
		vl = len(value)
		if vl < 5 or vl > 30:
			raise ValueError(
				f"Invalid SOAP action name length {vl} chars for "
				f"name '{value}'. Length must be >= 5 and <= 30."
			)
		return value

	def arguments_from(self, envelope: GenericEnvelope) -> _ArgsFrom:
		args = list()
		if self.body_type is not None:
			if envelope.body is None:
				raise f.MissingBodyFault(self.name)
			args.insert(self.body_index, envelope.body.content)
		if self.header_type is not None:
			if envelope.header is None:
				raise f.MissingHeaderFault(self.name)
			args.insert(self.header_index, envelope.header)
		return args

	def response_from(
			self,
			ret: t.Optional[MessageBody] = None,
			header: t.Optional[XroadHeader] = None
	) -> _RespFrom:
		has_return = self.return_type is not None
		if not has_return and ret is None:
			return Response()
		elif has_return and isinstance(ret, self.return_type):
			return SoapResponse(content=ret, header=header)
		raise TypeError(f"Expected return type {self.return_type}, but received {ret}")

	async def parse(self, http_body: bytes, content_type: str) -> GenericEnvelope:
		if content_type is None:
			raise f.InvalidContentTypeFault(content_type)
		body_type = content_type.split(';')[0]
		if body_type in ["text/xml", "application/xml", "application/soap+xml"]:
			return self.deserialize(http_body)
		elif body_type in ["multipart/related", "multipart/mixed"]:
			files: t.List[DecodedBodyPart] = list()
			decoder = MultipartDecoder(http_body, content_type)
			envelope = None
			for index, part in enumerate(decoder.parts):
				if index == 0:
					envelope = part
				elif part.is_mixed_multipart:
					content_type = part.headers.get('content-type')
					nested_decoder = MultipartDecoder(part.content, content_type)
					for nested_part in nested_decoder.parts:
						files.append(nested_part)
				else:
					files.append(part)
			http_body = self.process_files(envelope.content, files)
			return self.deserialize(http_body)
		raise f.InvalidContentTypeFault(body_type)

	def process_files(self, xml_str: bytes, files: t.List[DecodedBodyPart]) -> bytes:
		ids = [file.content_id for file in files]
		for cid in ids:
			if ids.count(cid) > 1:
				raise f.DuplicateCIDFault(cid)
		for file in files:
			cid = file.content_id.encode()
			count = xml_str.count(cid)
			if count == 0:
				raise f.MissingCIDFault(cid.decode())
			elif count > 1:
				raise f.DuplicateCIDFault(cid.decode())
			uid = self.storage.insert_object(file).encode()
			xml_str = xml_str.replace(cid, uid)
		return xml_str

	def deserialize(self, http_body: bytes) -> GenericEnvelope:
		if self.body_type is None:
			return EnvelopeFactory().deserialize(http_body)

		extra_nsmap = self.extract_extra_nsmap(http_body)
		attrs = self.body_type.__private_attributes__
		name = self.body_type.__name__
		last_err = None

		def inner(**kwargs) -> GenericEnvelope:
			strict_type = t.cast(t.Type[MessageBody], type(
				name, (self.body_type,), {}, extra="forbid", **kwargs
			))
			strict_type.__private_attributes__ = attrs
			envelope = EnvelopeFactory[strict_type]()
			return envelope.deserialize(http_body)

		if not extra_nsmap:
			return inner()
		for ns in extra_nsmap.keys():
			try:
				return inner(ns=ns, nsmap=extra_nsmap)
			except ValidationError as ex:
				last_err = ex
		raise last_err

	@staticmethod
	def extract_extra_nsmap(http_body: bytes) -> t.Dict[str, str]:
		parts = re.split(r'>\s*<', http_body.decode(), maxsplit=1)
		match = re.search(r'envelope', parts[0], re.IGNORECASE)
		if len(parts) == 1 or match is None:
			raise f.ClientFault("Unexpected envelope structure")
		parts = re.split(r' |xmlns:', parts[0])
		nsmap = dict()
		for part in parts:
			if '=' not in part:
				continue
			key, value = part.split('=')  # type: str, str
			if key not in HEADER_NSMAP:
				nsmap[key] = value.strip('"')
		return nsmap



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
from .binding import (
	SOAPFault,
	SOAPBody,
	SOAPHeader,
	WSDLFaultBinding,
	WSDLOutputBinding,
	WSDLInputBinding,
	XROADVersion,
	SOAPOperationBinding,
	WSDLOperationBinding,
	SOAPBinding,
	WSDLBinding
)
from .conditions import (
	Enumeration,
	RegexPattern,
	MinInclusive,
	MaxInclusive,
	TotalDigits,
	Length,
	MinLength,
	MaxLength,
	WhiteSpace,
)
from .definitions import (
	WSDLTypes,
	WSDLPart,
	WSDLMessage,
	WSDLDefinitions,
)
from .port_type import (
	WSDLFaultPort,
	WSDLOutputPort,
	WSDLInputPort,
	WSDLDocumentation,
	WSDLOperationPort,
	WSDLPortType,
)
from .restrictions import (
	NumericTypeRestriction,
	StringTypeRestriction
)
from .schema import (
	SimpleType,
	AnyXML,
	Element,
	Sequence,
	Attribute,
	ComplexType,
	Import,
	Include,
	Schema
)
from .service import (
	SOAPAddress,
	WSDLPortBinding,
	WSDLService
)


__all__ = [
	# binding.py
	"SOAPFault",
	"SOAPBody",
	"SOAPHeader",
	"WSDLFaultBinding",
	"WSDLOutputBinding",
	"WSDLInputBinding",
	"XROADVersion",
	"SOAPOperationBinding",
	"WSDLOperationBinding",
	"SOAPBinding",
	"WSDLBinding",

	# conditions.py
	"Enumeration",
	"RegexPattern",
	"MinInclusive",
	"MaxInclusive",
	"TotalDigits",
	"Length",
	"MinLength",
	"MaxLength",
	"WhiteSpace",

	# definitions.py
	"WSDLTypes",
	"WSDLPart",
	"WSDLMessage",
	"WSDLDefinitions",

	# port_type.py
	"WSDLFaultPort",
	"WSDLOutputPort",
	"WSDLInputPort",
	"WSDLDocumentation",
	"WSDLOperationPort",
	"WSDLPortType",

	# restrictions.py
	"NumericTypeRestriction",
	"StringTypeRestriction",

	# schema.py
	"SimpleType",
	"AnyXML",
	"Element",
	"Sequence",
	"Attribute",
	"ComplexType",
	"Import",
	"Include",
	"Schema",

	# service.py
	"SOAPAddress",
	"WSDLPortBinding",
	"WSDLService"
]



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
from .message_body import MessageBody, MessageBodyType
from .composite_meta import ElementSpecMeta, CompositeMeta
from .base_element_spec import BaseElementSpec
from .dynamic_spec import dynamic_spec


__all__ = [
	"MessageBody",
	"MessageBodyType",
	"ElementSpecMeta",
	"CompositeMeta",
	"BaseElementSpec",
	"dynamic_spec"
]



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from pydantic_xml import BaseXmlModel, attr
from .conditions import (
	Enumeration,
	RegexPattern,
	MinInclusive,
	MaxInclusive,
	TotalDigits,
	Length,
	MinLength,
	MaxLength,
	WhiteSpace
)


__all__ = [
	"NumericTypeRestriction",
	"StringTypeRestriction"
]


class NumericTypeRestriction(BaseXmlModel, tag="restriction"):
	base: str = attr()
	min_inclusive: t.Union[MinInclusive, None] = None
	max_inclusive: t.Union[MaxInclusive, None] = None
	total_digits: t.Union[TotalDigits, None] = None
	enumerations: t.Union[t.List[Enumeration], None] = None
	pattern: t.Union[RegexPattern, None] = None


class StringTypeRestriction(BaseXmlModel, tag="restriction"):
	base: str = attr()
	length: t.Union[Length, None] = None
	min_length: t.Union[MinLength, None] = None
	max_length: t.Union[MaxLength, None] = None
	enumerations: t.Union[t.List[Enumeration], None] = None
	whiteSpace: t.Union[WhiteSpace, None] = None
	pattern: t.Union[RegexPattern, None] = None



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
from .boolean import BooleanSpec, Boolean
from .date import DateSpec, Date
from .datetime import DateTimeSpec, DateTime
from .time import TimeSpec, Time
from .float import FloatSpec, Float
from .integer import IntegerSpec, Integer
from .string import StringSpec, String
from .netres import NetResSpec, NetRes
from .swaref import (
	SwaRefFile,
	SwaRefInternal,
	SwaRefSpec,
	SwaRefElement,
	SwaRef,
	SwaRefUtils
)


__all__ = [
	"BooleanSpec", "Boolean",
	"DateSpec", "Date",
	"DateTimeSpec", "DateTime",
	"TimeSpec", "Time",
	"FloatSpec", "Float",
	"IntegerSpec", "Integer",
	"StringSpec", "String",
	"NetResSpec", "NetRes",
	"SwaRefFile",
	"SwaRefInternal",
	"SwaRefSpec",
	"SwaRefElement",
	"SwaRef",
	"SwaRefUtils"
]



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import re
import chardet
import hashlib
import functools
import mimetypes
import typing as t
import charset_normalizer as charset
from pydantic import ValidationInfo
from ..uid_gen import UIDGenerator


__all__ = [
	"split_on_find",
	"guess_mime_type",
	"detect_decode",
	"convert_to_utf8",
	"remove_memory_addresses",
	"compute_signature",
	"is_incoming_request",
	"linearize_xml",
	"extract_content_ids"
]


_USB = t.Union[str, bytes]
_USN = t.Union[str, None]


def split_on_find(content: _USB, separator: _USB) -> t.Tuple[_USB, _USB]:
	point, sep_len = content.find(separator), len(separator)
	end = point if point != -1 else len(content) + 1
	start = end + sep_len if point != -1 else 0
	first, second = content[:end], content[start:]
	return first.strip(), second.strip()


def guess_mime_type(file_name: t.Union[str, None]) -> str:
	guess = mimetypes.guess_type(file_name)[0]
	default = "application/octet-stream"
	return guess or default


def detect_decode(string: bytes, encoding: str = 'utf-8') -> t.Tuple[_USB, _USN]:
	try:
		return string.decode(encoding), encoding
	except UnicodeDecodeError:
		if charset.is_binary(string):
			return string, None
		kwargs = dict(should_rename_legacy=True)
		match = chardet.detect(string, **kwargs)
		if match and match["confidence"] >= 0.7:  # pragma: no cover
			enc = match["encoding"]
			return string.decode(enc), enc
		return string, None  # pragma: no cover


def convert_to_utf8(string: bytes) -> bytes:
	decoded, encoding = detect_decode(string)
	if encoding not in [None, 'utf-8']:
		string = decoded.encode('utf-8')
	return string


def remove_memory_addresses(string: bytes) -> bytes:
	pattern = r"0x[0-9A-Fa-f]+"
	cleaned_text = re.sub(pattern, '', string.decode())
	return cleaned_text.encode()


@functools.lru_cache(maxsize=None)
def compute_signature(*args) -> str:
	repr_forms = []
	for arg in args:
		has_dict = hasattr(arg, "__dict__")
		subject = vars(arg) if has_dict else arg
		repr_forms.append(repr(subject))

	raw_sig = ''.join(repr_forms).encode()
	cleaned_sig = remove_memory_addresses(raw_sig)
	raw_digest = hashlib.shake_128(cleaned_sig).digest(12)
	return UIDGenerator(mode="key").generate(raw_digest)


def is_incoming_request(info: ValidationInfo) -> bool:
	return (info.context or {}).get("deserializing", False)


def linearize_xml(xml: str) -> bytes:
	return re.sub(
		pattern=r'>(\s+)<',
		string=xml.strip(),
		repl='><'
	).encode()


def extract_content_ids(xml_string: str, *, remove_from_xml: bool = False) -> t.Tuple[t.List[str], str]:
	matches = re.findall(r"(cid:[0-9a-fA-F]+)", xml_string)
	if remove_from_xml:  # pragma: no branch
		xml_string = re.sub(r"cid:[0-9a-fA-F]+", 'cid:', xml_string)
	return matches, xml_string



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import re
import typing as t
from collections import abc
from enum import Enum


__all__ = [
	"CommonValidators",
	"StringValidators",
	"NumberValidators"
]


class CommonValidators:
	min_value: t.Optional[t.Any] = None
	max_value: t.Optional[t.Any] = None
	pattern: t.Optional[str] = None
	enumerations: t.Optional[abc.Iterable[Enum]] = None

	@staticmethod
	def validate_type(obj: t.Any, expected_type: t.Any) -> None:
		if not isinstance(obj, expected_type):
			raise TypeError(
				f"invalid type {type(obj).__name__} for object "
				f"{obj}, expected type {expected_type}"
			)

	def validate_min_max_value(self, obj: t.Any) -> None:
		if self.min_value is not None and obj < self.min_value:
			raise ValueError(
				f"input value is less than the minimal value "
				f"of {self.min_value} $${obj}$$"
			)
		if self.max_value is not None and obj > self.max_value:
			raise ValueError(
				f"input value is greater than the maximum allowable "
				f"value of {self.max_value} $${obj}$$"
			)

	def validate_pattern(self, obj: t.Any) -> None:
		if self.pattern is None:
			return
		string = obj.isoformat() if hasattr(obj, "isoformat") else str(obj)
		match = re.search(self.pattern, string)
		if match is None:
			raise ValueError(
				f"input value does not match regex "
				f"pattern '{self.pattern}' $${string}$$"
			)

	def validate_enumerations(self, obj: t.Any) -> None:
		if self.enumerations is None:
			return
		values = [item.value for item in self.enumerations]
		if obj not in values:
			raise ValueError(
				f"input value is not one of the "
				f"allowed values: {values} $${obj}$$"
			)

	def validate_enum_value_types(self, expected_type: t.Type) -> None:
		if self.enumerations is None:
			return
		for item in self.enumerations:
			if not isinstance(item.value, expected_type):
				raise TypeError(
					f"enum value {item.value} is not of type {expected_type} "
					f"in enum {self.enumerations.__name__}"
				)


class StringValidators:
	length: t.Optional[int] = None
	min_length: t.Optional[int] = None
	max_length: t.Optional[int] = None
	whitespace: t.Literal["preserve", "replace", "collapse"] = "preserve"

	def validate_string_length(self, obj: str) -> None:
		string = str(obj)
		lstr = len(string)
		if self.length is not None and lstr != self.length:
			raise ValueError(
				f"input value length of {lstr} not equal "
				f"to {self.length} chars $${string}$$"
			)
		else:
			if self.min_length is not None and lstr < self.min_length:
				raise ValueError(
					f"input value length does not meet minimum requirement "
					f"of {self.min_length} chars $${string}$$"
				)
			if self.max_length is not None and lstr > self.max_length:
				raise ValueError(
					f"input value length exceeds maximum allowable "
					f"length of {self.max_length} chars $${string}$$"
				)

	def process_whitespace(self, obj: t.Any) -> str:
		if self.whitespace == "replace":
			return re.sub(r'\s', ' ', obj)
		elif self.whitespace == "collapse":
			return re.sub(r'\s+', ' ', obj)
		return obj


class NumberValidators:
	total_digits: t.Optional[int] = None

	def validate_total_digits(self, obj: t.Any) -> None:
		if not isinstance(obj, int):
			return
		if self.total_digits is not None:
			int_len = len(str(abs(obj)))
			if int_len > self.total_digits:
				raise ValueError(
					f"input value digits count exceeds the maximum allowable "
					f"count of {self.total_digits} digits $${obj}$$"
				)



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
import inflection
from ..constants import A8nType

try:
	from .message_body import MessageBody
except ImportError:  # pragma: no cover
	MessageBody: t.TypeAlias = t.Any
try:
	from .base_element_spec import BaseElementSpec
except ImportError:  # pragma: no cover
	BaseElementSpec: t.TypeAlias = t.Any


__all__ = [
	"validate_opt_mand_a8n",
	"validate_list_items",
	"validate_a8n_args"
]


def validate_opt_mand_a8n(attr: str, value: t.List[t.Any], spec: BaseElementSpec) -> None:
	elem_name = spec.tag or inflection.camelize(attr)
	count = len(value)

	if count > 1 and spec.a8n_type in [A8nType.MAND, A8nType.OPT]:
		raise ValueError(f"only one {elem_name} element is allowed, got {count}")
	elif count == 0 and spec.a8n_type == A8nType.MAND:
		raise ValueError(f"must provide at least one {elem_name} element")


def validate_list_items(attr: str, data: t.List[MessageBody], spec: BaseElementSpec):
	if not isinstance(data, list):
		raise ValueError(
			f"expected attribute '{attr}' value to be a list, "
			f"but received '{type(data).__name__}' instead."
		)
	count = len(data)
	if count < spec.min_occurs:
		raise ValueError(
			f"expected at least {spec.min_occurs} "
			f"{spec.tag} elements, got {count}"
		)
	elif isinstance(spec.max_occurs, int) and count > spec.max_occurs:
		raise ValueError(
			f"expected at most {spec.max_occurs} "
			f"{spec.tag} elements, got {count}"
		)
	expected = spec.internal_type or spec.element_type
	for item in data:
		if not isinstance(item, expected):
			raise ValueError(
				f"unexpected type '{type(item).__name__}' "
				f"in list for argument '{attr}'"
			)


def validate_a8n_args(a8n: t.Any, attr: str, cls_name: str, expected_type: t.Type) -> None:
	if a8n in [t.Optional, t.Union, t.List, list]:
		raise TypeError(
			f"Not permitted to provide '{a8n}' "
			f"annotation without a type argument."
		)
	if args := t.get_args(a8n):
		if len(args) == 1 and args[0] != expected_type:
			raise TypeError(
				f"Single annotation argument for class '{cls_name}' "
				f"attribute '{attr}' must be '{expected_type.__name__}'."
			)
		for arg in args:
			if arg is type(None) or arg == expected_type:
				continue
			arg_name = "None" if arg is type(None) else arg.__name__
			raise TypeError(
				f"Invalid annotation argument '{arg_name}' "
				f"for '{cls_name}' class attribute '{attr}'."
			)



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t
from ..constants import (
	ENV_NSMAP,
	XRO_NSMAP,
	IDEN_NSMAP
)
from .generics import (
	GenericEnvelope,
	GenericFault,
	GenericBody,
	AnyBody
)
from ..base import (
	MessageBodyType,
	MessageBody
)
from .header import XroadHeader


__all__ = ["EnvelopeFactory"]


class EnvelopeFactory(t.Generic[MessageBodyType]):
	_type: t.ClassVar[t.Type[MessageBody]] = None

	def __class_getitem__(cls, content_type: t.Type[MessageBody]):
		cls_name = f"{cls.__name__}[{content_type.__name__}]"
		if content_type.__xml_tag__ is None:
			content_type.__xml_tag__ = content_type.__name__
		return type(cls_name, (cls,), {"_type": content_type})

	def __init__(self, *, exclude_xroad_nsmap: bool = False) -> None:
		nsmap = {**ENV_NSMAP}
		if (
			not exclude_xroad_nsmap and self._type
			and not issubclass(self._type, GenericFault)
		):
			nsmap.update({**XRO_NSMAP, **IDEN_NSMAP})
			if isinstance(self._type.__xml_nsmap__, dict):
				nsmap.update(self._type.__xml_nsmap__)
		self._factory = t.cast(GenericEnvelope, type(
			'Factory', (GenericEnvelope,), {},
			ns="soapenv", nsmap=nsmap
		))

	def serialize(
			self,
			content: MessageBody,
			header: XroadHeader = None,
			pretty_print: bool = False,
			skip_empty: bool = False,
			encoding: str = 'utf-8'
	) -> bytes:
		if content.__class__ != self._type:
			raise TypeError(
				f"Cannot use '{content.__class__.__name__}' type instances "
				f"with '{self._type.__name__}' type envelope factory."
			)
		body_type = GenericBody[content.__class__]
		body = body_type(content=content)

		factory = self._factory[body_type]
		obj = factory(header=header, body=body)

		return obj.to_xml(
			standalone=header is None,
			pretty_print=pretty_print,
			skip_empty=skip_empty,
			encoding=encoding
		)

	def deserialize(self, content: t.Union[str, bytes]) -> GenericEnvelope:
		"""
		:param content: The XML string to be deserialized into an object.
		:raises pydantic.ValidationError: When the input XML string does
			not conform to the EnvelopeFactory subtype.
		:raises lxml.etree.LxmlError: When the input XML string contains
			syntax or structural errors.
		:return: An instance of MessageBody
		"""
		if not isinstance(content, bytes):
			content = content.encode("utf-8")

		model = AnyBody
		if self._type is not None:
			model = GenericBody[self._type]
		factory = self._factory[model]  # type: t.Type[MessageBody]
		obj = factory.from_xml(content, context={"deserializing": True})
		return t.cast(GenericEnvelope, obj)



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import typing as t


__all__ = [
	"get_annotations",
	"extract_parameter_positions"
]


def get_annotations(func: t.Any) -> dict:
	key: str = "__annotations__"
	anno = getattr(func, key, None)
	if anno is None and hasattr(func, "__dict__"):
		anno = func.__dict__.get(key, None)
	return anno or dict()


def extract_parameter_positions(anno: dict) -> dict:
	pos = dict(body=None, header=None)
	keys = list(anno.keys())
	for key in ["body", "header"]:
		if key in keys:
			pos[key] = keys.index(key)
	return pos



#
#   European Union Public License 1.2
#
#   Copyright (c) 2024, Centre of Registers and Information Systems
#
#   The contents of this file are subject to the terms and conditions defined in the License.
#   You may not use, modify, or distribute this file except in compliance with the License.
#
#   SPDX-License-Identifier: EUPL-1.2
#
import re
import typing as t
import inflection
from pydantic_xml import element
from ..base import MessageBody
from ..envelope import GenericFault
from .response import SoapResponse


__all__ = [
	"SoapFault",
	"ClientFault",
	"ServerFault",
	"InvalidMethodFault",
	"InvalidActionFault",
	"MissingActionFault",
	"InvalidContentTypeFault",
	"MissingBodyFault",
	"MissingHeaderFault",
	"MissingCIDFault",
	"DuplicateCIDFault",
	"ValidationFault"
]


class SoapFault(Exception):
	def __init__(
			self,
			code: str = "Client",
			string: str = "Bad Request",
			actor: t.Optional[str] = None,
			detail: t.Optional[MessageBody] = None,
			http_status_code: t.Optional[int] = 400
	) -> None:
		"""Base exception class of all exceptions raised by the library."""
		fault_type = MessageBody
		kwargs = dict(
			faultcode=code,
			faultstring=string,
			faultactor=actor
		)
		if isinstance(detail, MessageBody):
			fault_type = detail.__class__
			kwargs["detail"] = detail

		typed_fault = GenericFault[fault_type]
		fault = typed_fault(**kwargs)

		self.response = SoapResponse(
			http_status_code=http_status_code,
			content=fault,
			header=None
		)


class ClientFault(SoapFault):
	def __init__(self, ex: t.Union[str, Exception]) -> None:
		"""Exception that is raised when the client has made an invalid request."""
		super().__init__(string=str(ex))


class ServerFault(SoapFault):
	def __init__(self, ex: t.Union[str, Exception]) -> None:
		"""Error response that is returned to the client on uncaught exceptions."""
		super().__init__(
			http_status_code=500,
			string=str(ex),
			code="Server"
		)


class InvalidMethodFault(SoapFault):
	def __init__(self, method: str) -> None:
		msg = f"Service does not support {method} method for SOAP action requests"
		super().__init__(string=msg)


class InvalidActionFault(SoapFault):
	def __init__(self, action: str) -> None:
		msg = f"Service does not support SOAP action: {action}"
		super().__init__(string=msg)


class MissingActionFault(SoapFault):
	def __init__(self):
		msg = "SOAPAction HTTP header is missing"
		super().__init__(string=msg)


class InvalidContentTypeFault(SoapFault):
	def __init__(self, content_type: t.Any) -> None:
		msg = f"Invalid content type: {content_type}"
		super().__init__(string=msg)


class MissingBodyFault(SoapFault):
	def __init__(self, action_name: str) -> None:
		msg = f"Body element missing from envelope for SOAP action: {action_name}"
		super().__init__(string=msg)


class MissingHeaderFault(SoapFault):
	def __init__(self, action_name: str) -> None:
		msg = f"X-Road header element missing from envelope for SOAP action: {action_name}"
		super().__init__(string=msg)


class MissingCIDFault(SoapFault):
	def __init__(self, cid: str) -> None:
		msg = f"Content-ID missing from envelope: {cid}"
		super().__init__(string=msg)


class DuplicateCIDFault(SoapFault):
	def __init__(self, cid: str) -> None:
		msg = f"Duplicate Content-ID not allowed: {cid[4:]}"
		super().__init__(string=msg)


class ValidationFault(SoapFault):
	def __init__(self, ex: Exception) -> None:
		error = str(ex)
		parts = re.split(r'\n(?!\s\s)', error)
		match = re.search(r"^(.*?)\sfor", parts[0])
		string = match.group(1) if match else "validation error"
		super().__init__(
			http_status_code=400,
			detail=self.extract_details(parts),
			string=string + " in SOAP envelope",
			code="Client"
		)

	@classmethod
	def extract_details(cls, parts: t.List[str]) -> MessageBody:
		class Detail(MessageBody):
			location: str = element()
			reason: str = element()
			input_value: str = element(tag="inputValue")

		class ErrorDetails(MessageBody):
			details: t.List[Detail] = element(tag="validationError")

		details = list()
		for part in parts:
			loc, msg = None, part
			sub_parts = part.split('\n')
			if len(sub_parts) >= 2:
				loc, msg = sub_parts[:2]

			msg, input_value = cls.extract_input_value(msg)
			location = cls.extract_location(loc, msg)
			reason = cls.extract_reason(msg)

			params = [location, reason, input_value]
			if any(d != "Unknown" for d in params):
				details.append(Detail(
					location=location,
					reason=reason,
					input_value=input_value,
				))
		return ErrorDetails(details=details)

	@staticmethod
	def extract_input_value(msg: str) -> t.Tuple[str, str]:
		match = re.search(r"\$\$(.*?)\$\$", msg)
		if match:  # pragma: no cover
			msg = msg.replace(f"$${match.group(1)}$$", '')
		return msg, match.group(1) if match else "Unknown"

	@staticmethod
	def extract_reason(msg: str) -> str:
		match = re.search(r"\[line -?\d+]: (.+?) \[type=", msg)
		if match is None:  # pragma: no cover
			match = re.search(r"^\s\s(.+?) \[type=", msg)
		return match.group(1) if match else "Unknown"

	@staticmethod
	def extract_location(loc: str, msg: str) -> str:
		ln_match = re.search(r"(\[line -?\d+]: )", msg)
		return inflection.camelize(loc) if ln_match else "Unknown"
